'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (_ref) {
  var t = _ref.types;

  // Tests if the `name` is a valid import variable
  // or if it's a binding to a valid one.
  // Will return the name of the used variabel or null
  // if nothing was found
  function getUsedVariable(scope, name, importedVariable) {
    var binding = scope.getBinding(name);
    if (!binding) return null;
    if (t.isVariableDeclarator(binding.path.node) && binding.path.node.init) {
      return getUsedVariable(binding.path.scope, binding.path.node.init.name, importedVariable);
    }
    return importedVariable.has(name) && binding.kind === 'module' ? name : null;
  }

  return {
    pre: function pre() {
      // Used to track renaming imports in local file
      // eg. import { state as s } from 'cerebral/proxies';
      this.importedVariable = new Set();
    },

    visitor: {
      ImportDeclaration: function ImportDeclaration(path) {
        var importLocation = path.node.source.value;

        // The plan:
        //
        // 1) When import of cerebral/proxy found (directImport)
        //    a. Trigger conversion to tags
        //    b. Rewrite import to cerebral/tags

        // 2) When import containing cerebral.proxy in the name (indirectImport)
        //    a. Trigger conversion to tags

        var directImport = isDirectImport(importLocation); // 1)
        var indirectImport = isIndirectImport(importLocation); // 2)

        if (directImport || indirectImport) {
          // 1,2) a. Trigger conversion to tags

          // Complain about default import
          if (path.node.specifiers.some(function (item) {
            return t.isImportDefaultSpecifier(item);
          })) {
            throw path.buildCodeFrameError('Default import is not allowed');
          }

          // Verify that all imports are allowed and track the localName
          var _iteratorNormalCompletion = true;
          var _didIteratorError = false;
          var _iteratorError = undefined;

          try {
            for (var _iterator = path.node.specifiers[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
              var _step$value = _step.value,
                  importName = _step$value.imported.name,
                  localName = _step$value.local.name;

              if (isAllowedImport(importName)) {
                this.importedVariable.add(localName);
              } else if (directImport) {
                // Only complain about unkown imports
                // if it is a directImport
                throw path.buildCodeFrameError('"' + importName + '" is not a valid import');
              }
            }

            // 1) b. Rewrite import to cerebral/tags
          } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion && _iterator.return) {
                _iterator.return();
              }
            } finally {
              if (_didIteratorError) {
                throw _iteratorError;
              }
            }
          }

          if (directImport) {
            // Change import to the 'cerebral/tags'
            path.node.source.value = 'cerebral/tags';

            // Remove "path" from imports
            path.node.specifiers = path.node.specifiers.filter(function (item) {
              return item.imported.name !== 'statePath' && item.imported.name !== 'computedPath';
            });

            // remove if there aren't any specifiers left
            // This could happen if only `path` was used.
            if (path.node.specifiers.length === 0) {
              path.remove();
            }
          }
        }
      },
      MemberExpression: function MemberExpression(path) {
        // Always use the innermost MemberExpression
        if (!t.isIdentifier(path.node.object)) {
          return;
        }

        var tagName = getUsedVariable(path.scope, path.node.object.name, this.importedVariable);

        if (!tagName) {
          return;
        }

        var quasi = [];
        var quasis = [quasi];
        var expressions = [];
        var prevWasExpression = false;

        var rootMemberExpression = void 0;
        var currentMember = path;

        // Iterate trough all parents
        while (t.isMemberExpression(currentMember)) {
          var _currentMember = currentMember,
              _currentMember$node = _currentMember.node,
              property = _currentMember$node.property,
              computed = _currentMember$node.computed;

          // Plain id like state.a[1].b['test']

          if (isPlainPropertyAccess(t, property, computed)) {
            var value = t.isLiteral(property) ? property.value : property.name;
            quasi.push((quasi.length !== 0 || prevWasExpression ? '.' : '') + value);
            prevWasExpression = false;
            // Nested expressions like state.a[state.b]
          } else {
            quasi.push('.');
            expressions.push(property);
            quasi = [];
            quasis.push(quasi);
            prevWasExpression = true;
          }

          // Save the rootMember
          rootMemberExpression = currentMember;

          // Advance to next parent
          currentMember = currentMember.parentPath;
        }

        // Rewrite with a tagged template like
        // state`foo.bar`
        rootMemberExpression.replaceWith(t.taggedTemplateExpression(t.identifier(tagName), t.templateLiteral(quasis.map(function (v) {
          var str = v.join('');
          return t.templateElement({ raw: str, cooked: str });
        }), expressions)));
      }
    }
  };
};

var allowedImportPath = ['cerebral/proxy'];
var allowedImportVariables = ['props', 'state', 'signals', 'computed', 'moduleState', 'moduleComputed', 'moduleSignals'];

function isDirectImport(location) {
  return allowedImportPath.indexOf(location.toLowerCase()) >= 0;
}

function isIndirectImport(location) {
  return location.includes('cerebral.proxy');
}

function isAllowedImport(importName) {
  return allowedImportVariables.indexOf(importName) >= 0;
}

function isPlainPropertyAccess(t, property, computed) {
  return t.isIdentifier(property) && computed === false || t.isLiteral(property) && property.value;
}